'use strict';

const _ = require('lodash');
const co = require('co');

class Parameter {
    /**
     * A parameter to validate
     * @param {String} name - Name of the parameter. Can use dotted names.
     * @param {String} [location=body] - Location of the parameter
     */
    constructor(name, location) {
        if (!name) {
            throw new Error('Name is required');
        }
        this.name = name;
        this.location = location || 'body';
        this.isRequired = false;
        this.requiredErrorMessage = 'Required';
        this.validators = new Map();
    }
    
    /**
     * Set this parameter required
     * @param {String} [message] - Error message if this parameter is missing in the request
     * @returns {Parameter} - Returns the parameter object itself so that the it can be chained
     */
    required(message) {
        this.isRequired = true;
        this.requiredErrorMessage = message || this.requiredErrorMessage;
        return this;
    }
    
    /**
     * Set this parameter optional
     * @returns {Parameter} - Returns the parameter object itself so that the it can be chained
     */
    optional() {
        this.isRequired = false;
        return this;
    }
    
    /**
     * Validator Function
     * @callback Parameter~validatorCallback
     * @param {*} val - The value to validate
     * @this {Context} - Koa context object
     * @returns {Boolean|Promise.<Boolean>} - True for valid, false for invalid.
     */
    
    /**
     * Add a validation rule
     * @param {Parameter~validatorCallback} validator - A function to validate the parameter
     * @param {String} errorMessage - The error message to show if invalid.
     * @returns {Parameter} - Returns the parameter object itself so that the it can be chained
     */
    validate(validator, errorMessage) {
        if ('function' !== typeof validator) {
            throw new Error('Validator must be a function');
        }
        if (this.validators.has(errorMessage)) {
            throw new Error('Duplicate error message');
        }
        if (!errorMessage) {
            throw new Error('Error message missing');
        }
        this.validators.set(errorMessage, validator);
        return this;
    }
    
    /**
     * Check if this parameter is valid
     * @param {Context} ctx - The koa context object
     * @returns {Promise.<String|undefined>} - Returns a promise that
     * resolves to an error message if invalid or undefined if valid.
     */
    getErrorMessage(ctx) {
        var self = this;
        return co(function *() {
            var val = _.get(_.get(ctx.request, self.location), self.name);
            if (val) {
                for (var errMsg of self.validators.keys()) {
                    var validator = self.validators.get(errMsg);
                    var result = yield Promise.resolve(validator.call(ctx, val));
                    if (!result) {
                        return errMsg;
                    }
                }
            } else if (self.isRequired) {
                return self.requiredErrorMessage;
            }
            return undefined;
        });
    }
}

module.exports = Parameter;
