'use strict';

const Parameter = require('./parameter');
const co = require('co');
const _ = require('lodash');

class RequestValidator {
    /**
     * RequestValidator contains all the parameters to be validated
     */
    constructor() {
        this.parameters = [];
    }
    
    /**
     * Get or create a parameter
     * @param {String} name - Name of the parameter
     * @param {String} [location=body] - Location of the parameter
     * @returns {Parameter}
     */
    parameter(name, location) {
        location = location || 'body';
        for (var i = 0; i < this.parameters.length; i++) {
            var p = this.parameters[i];
            if (p.name === name && p.location === location) {
                return p;
            }
        }
        var param = new Parameter(name, location);
        this.parameters.push(param);
        return param;
    }
    
    /**
     * Check if the request in the context is valid
     * @param {Context} ctx - The koa context object
     * @returns {Promise.<Object>} - The error object
     */
    getError(ctx) {
        var self = this;
        return co(function *() {
            var errPromises = {};
            for (var i = 0; i < self.parameters.length; i++) {
                var param = self.parameters[i];
                errPromises[param.name] = param.getErrorMessage(ctx);
            }
            var errObj = yield errPromises;
            return _.pickBy(errObj, _.isString);
        });
    }
    
    /**
     * Get the koa middleware
     * @returns {Function}
     */
    middleware() {
        var self = this;
        return function *(next) {
            var err = yield self.getError(this);
            if (_.keys(err).length > 0) {
                this.body = err;
                this.response.status = 422;
            } else {
                yield next;
            }
        };
    }
}

module.exports = RequestValidator;
