/* eslint-env mocha */
'use strict';

const chai = require('chai');
chai.use(require('chai-as-promised'));
const expect = chai.expect;
const Parameter = require('../index').Parameter;

describe('Parameter', ()=> {
    describe('Constructor', ()=> {
        it('should return an instance of Parameter class', ()=> {
            var p = new Parameter('param');
            expect(p).to.be.an.instanceof(Parameter);
        });
        it('should not allow name arg to be null', ()=> {
            expect(()=> new Parameter()).to.throw('Name is required');
        });
        it('should set the location property if it is provided', ()=> {
            var p = new Parameter('param', 'somewhere');
            expect(p.location).to.eq('somewhere');
        });
        it('should set the location to "body" if it is missing', ()=> {
            var p = new Parameter('param');
            expect(p.location).to.eq('body');
        });
        it('should set the isRequired property to false', ()=> {
            var p = new Parameter('param');
            expect(p.isRequired).to.be.false;
        });
        it('should set the requiredErrorMessage property to "Required"', ()=> {
            var p = new Parameter('param');
            expect(p.requiredErrorMessage).to.eq('Required');
        });
    });
    describe('Instance Methods', ()=> {
        describe('required', ()=> {
            it('should set the isRequired property to true', ()=> {
                var p = new Parameter('param');
                p.required();
                expect(p.isRequired).to.be.true;
            });
            it('should also set the requiredErrorMessage if called with arg', ()=> {
                var p = new Parameter('param');
                p.required('new error message');
                expect(p.requiredErrorMessage).to.eq('new error message');
            });
            it('should return the parameter object itself', ()=> {
                var p = new Parameter('param');
                expect(p.required()).to.eq(p);
            });
        });
        describe('optional', ()=> {
            it('should set the isRequired property to false', ()=> {
                var p = new Parameter('param');
                p.required();
                expect(p.isRequired).to.be.true;
                p.optional();
                expect(p.isRequired).to.be.false;
            });
            it('should return the parameter object itself', ()=> {
                var p = new Parameter('param');
                expect(p.optional()).to.eq(p);
            });
        });
        describe('validate', ()=> {
            it('should add a new validation rule', ()=> {
                const SIZE_BEFORE_ADD = 0;
                const SIZE_AFTER_ADD = 1;
                const SIZE_AFTER_ANOTHER_ADD = 2;
                var p = new Parameter('param');
                expect(p.validators.size).to.eq(SIZE_BEFORE_ADD);
                p.validate(()=> true, 'err1');
                expect(p.validators.size).to.eq(SIZE_AFTER_ADD);
                p.validate(()=> false, 'err2');
                expect(p.validators.size).to.eq(SIZE_AFTER_ANOTHER_ADD);
            });
            it('should return the parameter object itself', ()=> {
                var p = new Parameter('param');
                expect(p.validate(()=> true, 'err')).to.eq(p);
            });
            it('should throw if the validator is not a function', ()=> {
                var p = new Parameter('param');
                // noinspection JSCheckFunctionSignatures
                expect(()=> p.validate('not function', 'err')).to.throw('Validator must be a function');
            });
            it('should throw if an error message already exists', ()=> {
                var p = new Parameter('param');
                p.validate(()=> true, 'err');
                expect(()=> p.validate(()=> true, 'err')).to.throw('Duplicate error message');
            });
            it('should throw if error message is missing', ()=> {
                var p = new Parameter('param');
                // noinspection JSCheckFunctionSignatures
                expect(()=> p.validate(()=> true)).to.throw('Error message missing');
            });
        });
        describe('getErrorMessage', ()=> {
            class Context {
            }
            var mockContext = new Context();
            mockContext.request = {
                body: {param: 1, dotted: {path: 2}},
                elsewhere: {a: 1}
            };
            it('should return a promise', ()=> {
                var p = new Parameter('param');
                expect(p.getErrorMessage(mockContext)).to.be.an.instanceof(Promise);
            });
            it('should return a promise that resolves to the error message if the validation fails', ()=> {
                var p = new Parameter('param');
                p.validate(val=> false, 'error message');
                expect(p.getErrorMessage(mockContext)).to.eventually.eq('error message');
            });
            it('should return a promise that resolves to undefined if the validation passes', ()=> {
                var p = new Parameter('param');
                p.validate(val=> true, 'error message');
                return expect(p.getErrorMessage(mockContext)).to.eventually.be.undefined;
            });
            it('should be ok for the validators to return promises', ()=> {
                var p = new Parameter('param');
                p.validate(val=> Promise.resolve(true), 'err1')
                    .validate(val=> Promise.resolve(false), 'err2');
                return expect(p.getErrorMessage(mockContext)).to.eventually.eq('err2');
            });
            it('should ran the validators in order and stops on the first failed rule', ()=> {
                var p = new Parameter('param');
                var validatorRanCount1 = 0;
                var validatorRanCount2 = 0;
                var validatorRanCount3 = 0;
                p
                    .validate(()=> {
                        validatorRanCount1++;
                        return true;
                    }, 'err1')
                    .validate(()=> {
                        validatorRanCount2++;
                        return false;
                    }, 'err2')
                    .validate(()=> {
                        validatorRanCount3++;
                        return false;
                    }, 'err3');
                p.getErrorMessage(mockContext)
                    .then(err=> {
                        expect(err).to.eq('err2');
                        expect(validatorRanCount1).to.eq(1);
                        expect(validatorRanCount2).to.eq(1);
                        expect(validatorRanCount3).to.eq(0);
                    });
            });
            it('should return a promise that rejects if a validator throws an error when called', ()=> {
                var p = new Parameter('param');
                p.validate(()=> {
                    throw new Error('error thrown');
                }, 'err msg');
                return expect(p.getErrorMessage(mockContext)).to.eventually.be.rejectedWith('error thrown');
            });
            it('should return a promise that rejects if the validator returns a rejected promise', ()=> {
                var p = new Parameter('param');
                p.validate(val=> Promise.reject(new Error('rejected promise')), 'err msg');
                return expect(p.getErrorMessage(mockContext)).to.eventually.be.rejectedWith('rejected promise');
            });
            it('should be ok to use dotted paths', ()=> {
                var p = new Parameter('dotted.path');
                p.validate(val=> val === 1, 'err');
                return expect(p.getErrorMessage(mockContext)).to.eventually.eq('err');
            });
            it('should look for the value to validate in the object specified by the "location" property', ()=> {
                var p = new Parameter('a', 'elsewhere');
                p.validate(val=> val !== 1, 'err');
                return expect(p.getErrorMessage(mockContext)).to.eventually.eq('err');
            });
            it('should fail the validation if the field is required but not present in the request', ()=> {
                var p = new Parameter('nonExistentParam');
                p.required();
                return expect(p.getErrorMessage(mockContext)).to.eventually.eq('Required');
            });
            it('should return a promise that resolves to the user defined error message for required ' +
                'if it was changed', ()=> {
                var p = new Parameter('b');
                p.required('another required error message');
                return expect(p.getErrorMessage(mockContext)).to.eventually.eq('another required error message');
            });
            it('should not fail the validation if the field is optional and the value is missing from the request, ' +
                'even if the value null/undefined would fail the validator call', ()=> {
                var p = new Parameter('b');
                p.optional()
                    .validate(val=> false, 'err');
                return expect(p.getErrorMessage(mockContext)).to.eventually.be.undefined;
            });
            it('should not run the validators if the field is optional and the value is missing ' +
                'from the request', ()=> {
                var p = new Parameter('b');
                var ranCount = 0;
                p.optional()
                    .validate(()=> {
                        ranCount++;
                        return false;
                    }, 'err');
                return p.getErrorMessage(mockContext)
                    .then(err=> {
                        expect(err).to.not.exist;
                        expect(ranCount).to.eq(0);
                    });
            });
        });
    });
});
