/* eslint-env mocha */
'use strict';

const chai = require('chai');
chai.use(require('chai-as-promised'));
const RequestValidator = require('../index').RequestValidator;
const Parameter = require('../index').Parameter;
const expect = chai.expect;
const koa = require('koa');
const rp = require('request-promise');

describe('RequestValidator', ()=> {
    describe('Constructor', ()=> {
        it('should return an instance of RequestValidator', ()=> {
            var rv = new RequestValidator();
            expect(rv).to.be.an.instanceof(RequestValidator);
        });
    });
    describe('Instance Methods', ()=> {
        describe('parameter', ()=> {
            it('should return an instance of the Field class', ()=> {
                var v = new RequestValidator();
                var param = v.parameter('path');
                expect(param).to.be.an.instanceof(Parameter);
            });
            it('should store and return a new parameter instance if path and where combination was' +
                ' not already there', ()=> {
                var v = new RequestValidator();
                var p1 = v.parameter('path1', 'where1');
                var p2 = v.parameter('path2', 'where1');
                var p3 = v.parameter('path2', 'where2');
                expect(p1).to.not.eq(p2);
                expect(p2).to.not.eq(p3);
                expect(p1).to.not.eq(p3);
            });
            it('should return the same parameter instance if path and where combination was already there', ()=> {
                var v = new RequestValidator();
                var p1 = v.parameter('path', 'where');
                var p2 = v.parameter('path', 'where');
                expect(p1).to.eq(p2);
            });
            it('should set the location arg to "body" if omitted', ()=> {
                var v = new RequestValidator();
                expect(v.parameter('path')).to.eq(v.parameter('path', 'body'));
                expect(v.parameter('path').location).to.eq('body');
            });
        });
        describe('getError', ()=> {
            class Context {
            }
            var mockContext = new Context();
            mockContext.request = {body: {a: 'a', b: 'b'}};
            it('should return a promise that resolves to an empty object if the request is valid', ()=> {
                var v = new RequestValidator();
                v.parameter('a').validate(val=> val === 'a', 'err a');
                v.parameter('b').validate(val=> val === 'b', 'err b');
                return expect(v.getError(mockContext)).to.eventually.deep.equal({});
            });
            it('should return a promise that resolves to an error object if the request is not valid', ()=> {
                var v = new RequestValidator();
                v.parameter('a').validate(val=> val !== 'a', 'err a');
                v.parameter('b').validate(val=> val !== 'b', 'err b');
                return expect(v.getError(mockContext)).to.eventually.deep.equal({a: 'err a', b: 'err b'});
            });
            it('should only contain the parameters that failed the validation in the error object', ()=> {
                var v = new RequestValidator();
                v.parameter('a').validate(val=> val !== 'a', 'err a');
                v.parameter('b').validate(val=> val === 'b', 'err b');
                return expect(v.getError(mockContext)).to.eventually.deep.equal({a: 'err a'});
            });
            it('should only have the first error message if multiple rules were not met', ()=> {
                var v = new RequestValidator();
                v.parameter('a')
                    .validate(val=> val !== 'a', 'err a1')
                    .validate(val=> val !== 'a', 'err a2')
                    .validate(val=> val !== 'a', 'err a3');
                v.parameter('b').validate(val=> val === 'b', 'err b');
                return expect(v.getError(mockContext)).to.eventually.deep.equal({a: 'err a1'});
            });
            it('should process the rules in order and breaks on the first unmet rule', ()=> {
                var v = new RequestValidator();
                var count1 = 0;
                var count2 = 0;
                var count3 = 0;
                v.parameter('a')
                    .validate(val=> {
                        count1++;
                        return val === 'a';
                    }, 'err1')
                    .validate(val=> {
                        count2++;
                        return val !== 'a';
                    }, 'err2')
                    .validate(val=> {
                        count3++;
                        return val !== 'a';
                    }, 'err3');
                return v.getError(mockContext)
                    .then(err=> {
                        expect(err).to.deep.equal({a: 'err2'});
                        expect(count1).to.eq(1);
                        expect(count2).to.eq(1);
                        expect(count3).to.eq(0);
                    });
            });
        });
        describe('middleware', ()=> {
            const PORT = 3000;
            var server;
            var app = koa();
            var router = require('koa-router')();
            
            var v1 = new RequestValidator();
            v1.parameter('val', 'header').required()
                .validate(val=> val === 'valid', 'invalid');
            router.get('/', v1.middleware(), function*() {
                this.body = {reachedHere: true};
            });
            
            var v2 = new RequestValidator();
            v2.parameter('val1', 'header').required();
            v2.parameter('val2', 'header').required()
                .validate(function (val2) {
                    return val2 !== this.request.header.val1;
                }, 'err');
            router.get('/this_arg', v2.middleware(), function*() {
                this.body = {reachedHere: true};
            });
            
            app.use(router.routes());
            app.use(router.allowedMethods());
            
            before(done=> {
                server = app.listen(PORT, done);
            });
            after(done=> {
                server.close(done);
            });
            
            var req = rp.defaults({baseUrl: 'http://localhost:3000/', json: true});
            it('should proceed to next handler in chain if request is valid', ()=>
                req({url: '/', headers: {val: 'valid'}})
                    .then(body=> {
                        expect(body).to.deep.equal({reachedHere: true});
                    })
            );
            it('should respond with 422 and an error object automatically if request is not valid', ()=>
                req({url: '/', headers: {val: 'invalid'}})
                    .then(()=> {
                        throw new Error('should not get here');
                    })
                    .catch(reason=> {
                        expect(reason.statusCode).to.eq(422);
                        expect(reason.error).to.deep.equal({val: 'invalid'});
                    })
            );
            it('should be able to use the "this" arg to refer to the context in the validation rule function', ()=>
                req({url: '/this_arg', headers: {val1: 'val', val2: 'val'}})
                    .then(()=> {
                        throw new Error('should not get here');
                    })
                    .catch(reason=> {
                        expect(reason.statusCode).to.eq(422);
                        expect(reason.error).to.deep.equal({val2: 'err'});
                    })
            );
        });
    });
});
